@@ -35,12 +35,18 @@
  * g_main_context_push_thread_default();
  */
 
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
 
 #include "gupnp-simple-igd.h"
 #include "gupnp-simple-igd-marshal.h"
 
-#include <string.h>
 #include <arpa/inet.h>
+#include <netinet/in.h>
+#include <string.h>
+#include <sys/types.h>
+#include <sys/socket.h>
 
 #include <libgupnp/gupnp.h>
 
@@ -48,6 +54,8 @@
 
 struct _GUPnPSimpleIgdPrivate
 {
+  GMainContext *main_context;
+
   GUPnPContextManager *gupnp_context_manager;
 
   GPtrArray *service_proxies;
@@ -96,11 +104,18 @@ enum
 {
   SIGNAL_MAPPED_EXTERNAL_PORT,
   SIGNAL_ERROR_MAPPING_PORT,
+  SIGNAL_CONTEXT_AVAILABLE,
   LAST_SIGNAL
 };
 
+/* props */
+enum
+{
+  PROP_0,
+  PROP_MAIN_CONTEXT
+};
 
-static guint signals[LAST_SIGNAL] = { 0 };
+guint signals[LAST_SIGNAL] = { 0 };
 
 
 #define GUPNP_SIMPLE_IGD_GET_PRIVATE(o)                                 \
@@ -114,6 +129,8 @@ G_DEFINE_TYPE (GUPnPSimpleIgd, gupnp_simple_igd, G_TYPE_OBJECT);
 static void gupnp_simple_igd_constructed (GObject *object);
 static void gupnp_simple_igd_dispose (GObject *object);
 static void gupnp_simple_igd_finalize (GObject *object);
+static void gupnp_simple_igd_get_property (GObject *object, guint prop_id,
+    GValue *value, GParamSpec *pspec);
 
 static void gupnp_simple_igd_gather (GUPnPSimpleIgd *self,
     struct Proxy *prox);
@@ -154,12 +171,20 @@ gupnp_simple_igd_class_init (GUPnPSimpleIgdClass *klass)
   gobject_class->constructed = gupnp_simple_igd_constructed;
   gobject_class->dispose = gupnp_simple_igd_dispose;
   gobject_class->finalize = gupnp_simple_igd_finalize;
+  gobject_class->get_property = gupnp_simple_igd_get_property;
 
   klass->add_port = gupnp_simple_igd_add_port_real;
   klass->remove_port = gupnp_simple_igd_remove_port_real;
 
+  g_object_class_install_property (gobject_class,
+      PROP_MAIN_CONTEXT,
+      g_param_spec_pointer ("main-context",
+          "The GMainContext to use",
+          "This GMainContext will be used for all async activities",
+          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+
   /**
-   * GUPnPSimpleIgd::mapped-external-port
+   * GUPnPSimpleIgd::mapped-external-port:
    * @self: #GUPnPSimpleIgd that emitted the signal
    * @proto: the requested protocol ("UDP" or "TCP")
    * @external_ip: the external IP
@@ -185,7 +210,7 @@ gupnp_simple_igd_class_init (GUPnPSimpleIgdClass *klass)
       G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING);
 
   /**
-   * GUPnPSimpleIgd::error-mapping-port
+   * GUPnPSimpleIgd::error-mapping-port:
    * @self: #GUPnPSimpleIgd that emitted the signal
    * @error: a #GError
    * @proto: The requested protocol
@@ -203,9 +228,31 @@ gupnp_simple_igd_class_init (GUPnPSimpleIgdClass *klass)
       0,
       NULL,
       NULL,
-      _gupnp_simple_igd_marshal_VOID__POINTER_STRING_UINT_STRING_UINT_STRING,
-      G_TYPE_NONE, 6, G_TYPE_POINTER, G_TYPE_STRING, G_TYPE_UINT,
+      _gupnp_simple_igd_marshal_VOID__BOXED_STRING_UINT_STRING_UINT_STRING,
+      G_TYPE_NONE, 6, G_TYPE_ERROR, G_TYPE_STRING, G_TYPE_UINT,
       G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING);
+
+  /**
+   * GUPnPSimpleIgd::context-available:
+   * @self: #GUPnPSimpleIgd that emitted the signal
+   * @context: a #GUPnPContext
+   *
+   * This is to allow the application to control which #GUPnPContext this
+   * client should use. If the application connects to this signal, it controls
+   * if a context will be used by changing the return value of the signal
+   * handler.
+   *
+   * Returns: FALSE if the context should be used or TRUE if it should
+   * be ignored
+   */
+  signals[SIGNAL_CONTEXT_AVAILABLE] = g_signal_new ("context-available",
+      G_TYPE_FROM_CLASS (klass),
+      G_SIGNAL_RUN_LAST,
+      0,
+      NULL,
+      NULL,
+      _gupnp_simple_igd_marshal_BOOLEAN__OBJECT,
+      G_TYPE_BOOLEAN, 1, G_TYPE_OBJECT);
 }
 
 static void
@@ -403,12 +450,32 @@ gupnp_simple_igd_finalize (GObject *object)
 {
   GUPnPSimpleIgd *self = GUPNP_SIMPLE_IGD_CAST (object);
 
+  g_main_context_unref (self->priv->main_context);
+
   g_warn_if_fail (self->priv->mappings->len == 0);
   g_ptr_array_free (self->priv->mappings, TRUE);
 
   G_OBJECT_CLASS (gupnp_simple_igd_parent_class)->finalize (object);
 }
 
+static void
+gupnp_simple_igd_get_property (GObject *object, guint prop_id,
+    GValue *value, GParamSpec *pspec)
+{
+  GUPnPSimpleIgd *self = GUPNP_SIMPLE_IGD_CAST (object);
+
+  switch (prop_id) {
+    case PROP_MAIN_CONTEXT:
+      g_value_set_pointer (value, self->priv->main_context);
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+  }
+
+}
+
+
 static void
 _cp_service_avail (GUPnPControlPoint *cp,
     GUPnPServiceProxy *proxy,
@@ -467,6 +534,8 @@ gupnp_simple_igd_add_control_point (GUPnPSimpleIgd *self,
 
   cp = gupnp_control_point_new (gupnp_context, target);
   g_return_if_fail (cp);
+  g_assert (GUPNP_IS_CONTROL_POINT (cp));
+  g_assert (G_IS_OBJECT (self));
 
   g_signal_connect_object (cp, "service-proxy-available",
       G_CALLBACK (_cp_service_avail), self, 0);
@@ -485,6 +554,13 @@ _context_available (GUPnPContextManager *manager, GUPnPContext *gupnp_context,
     GUPnPSimpleIgd *self)
 {
   SoupSession *session;
+  gboolean ignore_context = FALSE;
+
+  g_signal_emit (self, signals[SIGNAL_CONTEXT_AVAILABLE], 0, gupnp_context,
+      &ignore_context);
+
+  if (ignore_context)
+    return;
 
   session = gupnp_context_get_session (gupnp_context);
   g_object_set (session, "timeout", SOUP_REQUEST_TIMEOUT, NULL);
@@ -502,10 +578,15 @@ gupnp_simple_igd_constructed (GObject *object)
   GUPnPSimpleIgd *self = GUPNP_SIMPLE_IGD_CAST (object);
   SoupSession *session;
 
+  self->priv->main_context = g_main_context_get_thread_default ();
+  if (!self->priv->main_context)
+    self->priv->main_context = g_main_context_default ();
+  g_main_context_ref (self->priv->main_context);
+
   self->priv->gupnp_context_manager = gupnp_context_manager_create (0);
 
-  g_signal_connect (self->priv->gupnp_context_manager, "context-available",
-      G_CALLBACK (_context_available), self);
+  g_signal_connect_object (self->priv->gupnp_context_manager,
+      "context-available", G_CALLBACK (_context_available), self, 0);
 
   if (G_OBJECT_CLASS (gupnp_simple_igd_parent_class)->constructed)
     G_OBJECT_CLASS (gupnp_simple_igd_parent_class)->constructed (object);
@@ -710,7 +791,7 @@ _service_proxy_added_port_mapping (GUPnPServiceProxy *proxy,
         g_timeout_source_new_seconds (pm->mapping->lease_duration / 2);
       g_source_set_callback (pm->renew_src,
           _renew_mapping_timeout, pm, NULL);
-      g_source_attach (pm->renew_src, g_main_context_get_thread_default ());
+      g_source_attach (pm->renew_src, self->priv->main_context);
     }
   }
   else
